# -*- coding: utf-8 -*-
#
# Author: Natalia B. Bidart <natalia.bidart@canonical.com>
#
# Copyright (C) 2009 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the protocol client."""

import unittest
import uuid

from twisted.internet.defer import Deferred

from ubuntuone.storageprotocol import protocol_pb2, sharersp
from ubuntuone.storageprotocol.client import (
    StorageClient, CreateUDF, ListVolumes, DeleteVolume
)
from ubuntuone.storageprotocol import volumes

PATH = u'~/Documents/pdfs/moño/'
NAME = u'UDF-me'
VOLUME = uuid.UUID('12345678-1234-1234-1234-123456789abc')
NODE = uuid.UUID('FEDCBA98-7654-3211-2345-6789ABCDEF12')
USER = u'Dude'

class FakedError(Exception):
    """Stub to replace Request.error."""


def stub_function(*args, **kwargs):
    """Stub to replace non-related functions."""
    return None

def faked_error(message):
    """Stub to replace Request.error."""
    raise FakedError

def was_called(self, flag):
    """Helper to assert a function was called."""
    assert not getattr(self, flag)
    def set_flag(*args, **kwargs):
        """Record the calling was made."""
        setattr(self, flag, True)
    return set_flag

def build_list_volumes():
    """Build a LIST_VOLUMES message."""
    message = protocol_pb2.Message()
    message.type = protocol_pb2.Message.VOLUMES_INFO
    return message

def build_volume_created():
    """Build VOLUME_CREATED message."""
    message = protocol_pb2.Message()
    message.type = protocol_pb2.Message.VOLUME_CREATED
    return message

def build_volume_deleted():
    """Build VOLUME_DELETED message."""
    message = protocol_pb2.Message()
    message.type = protocol_pb2.Message.VOLUME_DELETED
    return message

def set_root_message(message):
    """Set a simple Root message."""
    message.type = protocol_pb2.Volumes.ROOT
    message.root.node = str(NODE)

def set_udf_message(message):
    """Set a simple UDFs message."""
    message.type = protocol_pb2.Volumes.UDF
    message.udf.volume = str(VOLUME)
    message.udf.node = str(NODE)
    message.udf.suggested_path = PATH

def set_share_message(message):
    """Set a simple Share message."""
    message.type = protocol_pb2.Volumes.SHARE
    message.share.share_id = str(VOLUME)
    message.share.direction = 0
    message.share.subtree = str(NODE)
    message.share.share_name = u'test'
    message.share.other_username = USER
    message.share.other_visible_name = USER
    message.share.accepted = False
    message.share.access_level = 0


class DummyAttribute(object):
    """Helper class to replace non-related classes."""

    def __getattribute__(self, name):
        """Attributes can be whatever we need."""
        return stub_function


class FakedProtocol(StorageClient):
    """Fake StorageClient to avoid twisted."""

    def __init__(self, *args, **kwargs):
        """Override transports and keep track of messages."""
        StorageClient.__init__(self, *args, **kwargs)
        self.transport = DummyAttribute()
        self.messages = []

    def sendMessage(self, message):
        """Keep track of messages."""
        self.messages.append(message)


class ClientTestCase(unittest.TestCase):
    """Check that MultiQuery works using an iterator."""

    def setUp(self):
        """Initialize testing client."""
        self.client = FakedProtocol()
        self.called = False
        self.volume = None

    def tearDown(self):
        """Clean up."""
        self.client = None

    # client to server
    def test_create_udf(self):
        """Test create_udf."""
        original = CreateUDF.start
        CreateUDF.start = was_called(self, 'called')

        try:
            result = self.client.create_udf(path=PATH, name=NAME)
            self.assertTrue(self.called, 'CreateUDF.start() was called')
            self.assertTrue(isinstance(result, Deferred))
        finally:
            CreateUDF.start = original

    def test_list_volumes(self):
        """Test list_volumes."""
        original = ListVolumes.start
        ListVolumes.start = was_called(self, 'called')

        try:
            result = self.client.list_volumes()
            self.assertTrue(self.called, 'ListVolumes.start() was called')
            self.assertTrue(isinstance(result, Deferred))
        finally:
            ListVolumes.start = original

    def test_delete_volume(self):
        """Test delete_volume."""
        original = DeleteVolume.start
        DeleteVolume.start = was_called(self, 'called')

        try:
            result = self.client.delete_volume(volume_id=VOLUME)
            self.assertTrue(self.called, 'DeleteVolume.start() was called')
            self.assertTrue(isinstance(result, Deferred))
        finally:
            DeleteVolume.start = original

    def test_set_volume_deleted_callback(self):
        """Test callback setting."""
        a_callback = lambda x: None
        self.client.set_volume_deleted_callback(a_callback)
        # pylint: disable-msg= W0212
        self.assertTrue(self.client._volume_deleted_callback is a_callback)

    def test_callback_must_be_callable(self):
        """Test set callback parameters."""
        self.assertRaises(TypeError, self.client.set_volume_created_callback,
                          'hello')

        self.assertRaises(TypeError, self.client.set_volume_deleted_callback,
                          'world')

    def test_set_volume_created_callback(self):
        """Test callback setting."""
        a_callback = lambda y, z: None
        self.client.set_volume_created_callback(a_callback)
        # pylint: disable-msg= W0212
        self.assertTrue(self.client._volume_created_callback is a_callback)

    # share notification callbacks
    def test_share_change_callback(self):
        """Test share_change callback usage."""
        self.assertRaises(TypeError, self.client.set_share_change_callback,
                          'hello')
        #create a response and message
        share_resp = sharersp.NotifyShareHolder.from_params(
           uuid.uuid4(), uuid.uuid4(), 'sname', 'byu', 'tou', 'View')
        proto_msg = protocol_pb2.Message()
        proto_msg.type = protocol_pb2.Message.NOTIFY_SHARE
        share_resp.dump_to_msg(proto_msg.notify_share)

        #wire up a call back and make sure it's correct
        self.share_notif = None
        a_callback = lambda notif: setattr(self, 'share_notif', notif)
        self.client.set_share_change_callback(a_callback)
        self.assertTrue(self.client._share_change_callback is a_callback)
        self.client.handle_NOTIFY_SHARE(proto_msg)
        self.assertEquals(self.share_notif.share_id, share_resp.share_id)

    def test_share_delete_callback(self):
        """Test share_delete callback usage."""
        self.assertRaises(TypeError, self.client.set_share_delete_callback,
                          'hello')

        share_id = uuid.uuid4()
        proto_msg = protocol_pb2.Message()
        proto_msg.type = protocol_pb2.Message.SHARE_DELETED
        proto_msg.share_deleted.share_id = str(share_id)

        #wire up a call back and make sure it's correct
        self.deleted_share_id = None
        a_callback = lambda notif: setattr(self, 'deleted_share_id', notif)
        self.client.set_share_delete_callback(a_callback)
        self.assertTrue(self.client._share_delete_callback is a_callback)
        self.client.handle_SHARE_DELETED(proto_msg)
        self.assertEquals(self.deleted_share_id, share_id)

    def test_share_answer_callback(self):
        """Test share_answer callback usage."""
        self.assertRaises(TypeError, self.client.set_share_answer_callback,
                          'hello')

        share_id = uuid.uuid4()
        proto_msg = protocol_pb2.Message()
        proto_msg.type = protocol_pb2.Message.SHARE_ACCEPTED
        proto_msg.share_accepted.share_id = str(share_id)
        proto_msg.share_accepted.answer = protocol_pb2.ShareAccepted.YES

        #wire up a call back and make sure it's correct
        self.answer = None
        a_callback = lambda s, a: setattr(self, 'answer', (s, a))
        self.client.set_share_answer_callback(a_callback)
        self.assertTrue(self.client._share_answer_callback is a_callback)
        self.client.handle_SHARE_ACCEPTED(proto_msg)
        self.assertEquals(self.answer[0], share_id)
        self.assertEquals(self.answer[1], "Yes")

    # server to client
    def test_handle_volume_created(self):
        """Test handle_VOLUME_CREATED."""
        a_callback = was_called(self, 'called')
        self.client.set_volume_created_callback(a_callback)

        message = build_volume_created()
        set_root_message(message.volume_created)
        self.client.handle_VOLUME_CREATED(message)

        self.assertTrue(self.called)

    def test_handle_root_created_passes_a_root(self):
        """Test handle_VOLUME_CREATED parameter passing."""
        self.volume = None
        a_callback = lambda vol: setattr(self, 'volume', vol)
        self.client.set_volume_created_callback(a_callback)

        message = build_volume_created()
        set_root_message(message.volume_created)
        root = volumes.RootVolume.from_msg(message.volume_created.root)

        self.client.handle_VOLUME_CREATED(message)
        self.assertEquals(root, self.volume)

    def test_handle_udf_created_passes_a_udf(self):
        """Test handle_VOLUME_CREATED parameter passing."""
        self.volume = None
        a_callback = lambda vol: setattr(self, 'volume', vol)
        self.client.set_volume_created_callback(a_callback)

        message = build_volume_created()
        set_udf_message(message.volume_created)
        udf = volumes.UDFVolume.from_msg(message.volume_created.udf)

        self.client.handle_VOLUME_CREATED(message)
        self.assertEquals(udf, self.volume)

    def test_handle_share_created_passes_a_share(self):
        """Test handle_VOLUME_CREATED parameter passing."""
        self.volume = None
        a_callback = lambda vol: setattr(self, 'volume', vol)
        self.client.set_volume_created_callback(a_callback)

        message = build_volume_created()
        set_share_message(message.volume_created)
        share = volumes.ShareVolume.from_msg(message.volume_created.share)

        self.client.handle_VOLUME_CREATED(message)
        self.assertEquals(share, self.volume)

    def test_handle_volume_created_if_volume_is_buggy(self):
        """Test handle_VOLUME_CREATED if the volume is buggy."""
        message = build_volume_created()
        message.volume_created.type = -1 # invalid type!
        self.client.set_volume_created_callback(lambda vol: None)
        self.assertRaises(TypeError, self.client.handle_VOLUME_CREATED,
                          message)

    def test_handle_volume_created_if_callback_is_none(self):
        """Test handle_VOLUME_CREATED if callback is none."""
        message = build_volume_created()
        self.client.handle_VOLUME_CREATED(message)

    def test_handle_volume_deleted(self):
        """Test handle_VOLUME_DELETED."""
        a_callback = was_called(self, 'called')
        self.client.set_volume_deleted_callback(a_callback)

        message = build_volume_deleted()
        message.volume_deleted.volume = str(VOLUME)
        self.client.handle_VOLUME_DELETED(message)

        self.assertTrue(self.called)

    def test_handle_volume_deleted_passes_the_id(self):
        """Test handle_VOLUME_DELETED."""
        self.volume = None
        a_callback = lambda vol_id: setattr(self, 'volume', vol_id)
        self.client.set_volume_deleted_callback(a_callback)

        message = build_volume_deleted()
        message.volume_deleted.volume = str(VOLUME)
        self.client.handle_VOLUME_DELETED(message)

        self.assertEquals(VOLUME, self.volume)

    def test_handle_volume_deleted_if_none(self):
        """Test handle_VOLUME_DELETED if callback is none."""
        message = build_volume_deleted()
        self.client.handle_VOLUME_DELETED(message)


class CreateUDFTestCase(unittest.TestCase):
    """Test cases for CreateUDF op."""

    def setUp(self):
        """Initialize testing protocol."""
        self.protocol = FakedProtocol()
        self.request = CreateUDF(self.protocol, path=PATH, name=NAME)
        self.request.error = faked_error
        self.done_called = False
        self.request.done = was_called(self, 'done_called')

    def tearDown(self):
        """Clean up."""
        self.protocol = None
        self.request = None

    def test_init(self):
        """Test request creation."""
        self.assertEquals(PATH, self.request.path)
        self.assertEquals(NAME, self.request.name)
        self.assertTrue(self.request.volume_id is None)
        self.assertTrue(self.request.node_id is None)

    def test_start(self):
        """Test request start."""
        self.request.start()

        self.assertEquals(1, len(self.request.protocol.messages))
        actual_msg, = self.request.protocol.messages
        self.assertEquals(protocol_pb2.Message.CREATE_UDF, actual_msg.type)
        self.assertEquals(self.request.path, actual_msg.create_udf.path)
        self.assertEquals(self.request.name, actual_msg.create_udf.name)

    def test_process_message_error(self):
        """Test request processMessage on error."""
        message = protocol_pb2.Message()
        self.assertRaises(FakedError, self.request.processMessage, message)

    def test_process_message_volume_created(self):
        """Test request processMessage on sucess."""
        message = protocol_pb2.Message()
        message.type = protocol_pb2.Message.VOLUME_CREATED
        message.volume_created.type = protocol_pb2.Volumes.UDF
        message.volume_created.udf.volume = str(VOLUME)
        message.volume_created.udf.node = str(NODE)
        self.request.processMessage(message)

        self.assertEquals(str(VOLUME), self.request.volume_id, 'volume set')
        self.assertEquals(str(NODE), self.request.node_id, 'node set')
        self.assertTrue(self.done_called, 'done() was called')


class ListVolumesTestCase(unittest.TestCase):
    """Test cases for ListVolumes op."""

    def setUp(self):
        """Initialize testing protocol."""
        self.protocol = FakedProtocol()
        self.request = ListVolumes(self.protocol)
        self.request.error = faked_error
        self.done_called = False
        self.request.done = was_called(self, 'done_called')

    def tearDown(self):
        """Clean up."""
        self.protocol = None
        self.request = None

    def test_init(self):
        """Test request creation."""
        self.assertEquals([], self.request.volumes)

    def test_start(self):
        """Test request start."""
        self.request.start()

        self.assertEquals(1, len(self.request.protocol.messages))
        actual_msg, = self.request.protocol.messages
        self.assertEquals(protocol_pb2.Message.LIST_VOLUMES, actual_msg.type)

    def test_process_message_error(self):
        """Test request processMessage on error."""
        message = protocol_pb2.Message()
        self.assertRaises(FakedError, self.request.processMessage, message)

    def test_process_message_error_when_incorrect_volume(self):
        """Test error condition when incorrect volume type."""
        message = build_list_volumes()
        message.list_volumes.type = -1
        self.assertRaises(FakedError, self.request.processMessage, message)

    def test_process_message_volume_created(self):
        """Test request processMessage on sucess."""
        message = build_list_volumes()
        set_udf_message(message.list_volumes)
        udf = volumes.UDFVolume.from_msg(message.list_volumes.udf)
        self.request.processMessage(message)

        message = build_list_volumes()
        set_share_message(message.list_volumes)
        share = volumes.ShareVolume.from_msg(message.list_volumes.share)
        self.request.processMessage(message)

        message = build_list_volumes()
        set_root_message(message.list_volumes)
        root = volumes.RootVolume.from_msg(message.list_volumes.root)
        self.request.processMessage(message)

        self.assertEquals(3, len(self.request.volumes),
                          '3 volumes stored')
        self.assertEquals([udf, share, root], self.request.volumes,
                          'volumes stored')

        message = protocol_pb2.Message()
        message.type = protocol_pb2.Message.VOLUMES_END
        self.request.processMessage(message)

        self.assertTrue(self.done_called, 'done() was called')

    def test_start_cleanups_volumes(self):
        """Test start() is idempotent."""
        self.request.start()

        message = build_list_volumes()
        set_udf_message(message.list_volumes)
        self.request.processMessage(message)

        message = protocol_pb2.Message()
        message.type = protocol_pb2.Message.VOLUMES_END
        self.request.processMessage(message)

        self.request.start()
        self.assertEquals([], self.request.volumes)


class DeleteVolumeTestCase(unittest.TestCase):
    """Test cases for DeleteVolume op."""

    def setUp(self):
        """Initialize testing protocol."""
        self.protocol = FakedProtocol()
        self.request = DeleteVolume(self.protocol, volume_id=VOLUME)
        self.request.error = faked_error
        self.done_called = False
        self.request.done = was_called(self, 'done_called')

    def tearDown(self):
        """Clean up."""
        self.protocol = None
        self.request = None

    def test_init(self):
        """Test request creation."""
        self.assertEquals(str(VOLUME), self.request.volume_id)

    def test_start(self):
        """Test request start."""
        self.request.start()

        self.assertEquals(1, len(self.request.protocol.messages))
        actual_msg, = self.request.protocol.messages
        self.assertEquals(protocol_pb2.Message.DELETE_VOLUME, actual_msg.type)
        self.assertEquals(self.request.volume_id,
                          actual_msg.delete_volume.volume)

    def test_process_message_error(self):
        """Test request processMessage on error."""
        message = protocol_pb2.Message()
        self.assertRaises(FakedError, self.request.processMessage, message)

    def test_process_message_ok(self):
        """Test request processMessage on sucess."""
        message = protocol_pb2.Message()
        message.type = protocol_pb2.Message.OK
        self.request.processMessage(message)

        self.assertTrue(self.done_called, 'done() was called')


if __name__ == '__main__':
    unittest.main()
